/*****************************************************************************
* CINEDB.CPP                                                                 *
*                                                                            *
* Cinematronics Emulator / Debugger.                                         *
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <pitrex/pitrexio-gpio.h>
#include <vectrex/vectrexInterface.h>
#include <vectrex/osWrapper.h>

#define   VERSION   "Version 1.3"

#include "cines.h"

int filelength(FILE *fd)		// also used in inifile.c so not static
{
   int fsize;
   fseek (fd, 0, SEEK_END);
// my book says return 0 on success!
   fsize = ftell(fd); 
   (void) fseek (fd, 0, SEEK_SET);
   return fsize;
}

/*****************************************************************************
 * Read ROM images.                                                          *
 *                                                                           *
 * Returns:                                                                  *
 *      0 - Ok                                                               *
 *      1 - Error, (error message already displayed)                         *
 *****************************************************************************/
uint readRoms (char *romName[8], uchar OptMemSize)
{
   FILE *inFile;
   char *aFileName;
   int ii;
   int base[2];
   int addr;
   unsigned int sizeRead;
   uint romSize=0, romCount=0, evnOdd;
   ulong tempSize;

   switch (OptMemSize) {
   case 0:			// 4k games
      romSize = 2048;      romCount = 2;
      break;

   case 1:			// 8k games
      romSize = 2048;      romCount = 4;
      break;

   case 2:			// 16k games
      romSize = 4096;      romCount = 4;
      break;

   case 3:			// 32k game
      romSize = 4096;      romCount = 8;
      break;

   case 4:			// 32k game
      romSize = 8192;      romCount = 4;
      break;
   }

   evnOdd = 0;			// zero even odd index
   base[0] = 0x0000; base[1] = 0x0001;
   for (ii = 0; ii < romCount; ii++) {
     addr = base[evnOdd];
     aFileName = romName[ii];

     inFile = fopen (aFileName, "r");
     if (inFile == 0) {
         printf ("readRoms: error opening file: \"%s\"\n", aFileName);
	 return (1);
      }
      tempSize = (int) filelength (inFile);	// get length of file BEFORE ANY READING IS DONE

      if (tempSize != romSize) {
	fclose (inFile);
        printf ("readRoms: file \"%s\" wrong size. (Must be %u bytes.)\n", aFileName, romSize);
        return (1);
      }
      fprintf(stderr, "Loading %s at 0x%04x\n", aFileName, addr);

      sizeRead = 0;
      for (;;) {
        int c = fgetc(inFile);
        if (c == EOF) break;
	rom[addr] = c;
	addr += 2; base[evnOdd] += 2;
	sizeRead += 1;
      }
      if (sizeRead != romSize)
      {
        fclose (inFile);
        printf ("readRoms: error reading file: \"%s\"(%i,%i)\r\n", aFileName, sizeRead, romSize);
        return (1);
      }
      fclose (inFile);

      evnOdd ^= 0x0001;
   }
   return (0);
}


/**************************** main **********************************/
int Game;

#define No 0
#define Yes 1

int main (int argc, char **argv)
{
   int Rotate;
   int Flip_X;
   int Flip_Y;
   int Switches;
   int Inputs;
   int MemSize=0;
   int JMI=0;
   char *RomImages[8];
   static int parity = 0;

   Game = 0; useParity = 0;

   vectrexinit (1);
   v_init ();
   usePipeline = 1;


#ifdef FREESTANDING
   char *_argv[3];
   _argv[0] = "cine.img";
   _argv[1] = getLoadParameter ();
   _argv[2] = NULL;
   argc = 2;
   argv = _argv;
   printf ("Game selected: %s\r\n", _argv[1]);
#endif

   static char hdr[] =
    "CINEMU - Cinematronics Emulator. " VERSION "\n"
    "(c) Copyright 1997, Zonn Moore.  All rights reserved.\n";

   if ((argc == 2) && (strcmp(argv[1], "-V") == 0)) {
     fputs (hdr, stderr);
     exit(0);
   }

   if ((argc == 3) && (strcmp(argv[1], "-d") == 0)) {
     argc -= 1; argv += 1; ccpudebug = 1;
   }

   if ((argc != 2) || (strcmp(argv[1], "-h") == 0)) {
      fprintf (stderr, "syntax: cinemu [-h | -V | [-d] <game>]  where the supported games are:\n");
      fprintf (stderr, "  armorattack  boxingbugs   demon  ripoff      spacewars   starcastle  sundance    waroftheworlds\n");
      fprintf (stderr, "  barrier      cosmicchasm  qb3    solarquest  speedfreak  starhawk    tailgunner  warrior\n");
      fprintf (stderr, "\n");
      exit (1);
   }
   /*
   * Games known to work 100% or nearly so, at some point or another of
   * the emulators development (but not necessarily right now :):
   * RipOff, Solar Quest, Spacewar, Speed Freak, Star Castle, Star Hawk,
   * Tail Gunner, War of the Worlds, Warrior
   *
   * For reference, all of the cinematronics games are:
   * Armor Attack, Barrier, Boxing Bugs, Demon, Ripoff, Solar Quest,
   * Spacewar, Speed Freak, Star Castle, Star Hawk, Sundance, Tail Gunner
   * War of the worlds, Warrior
   */

   // It's looking like all of the cinematronics games need the 'parity' tweak to handle 2 consecutive WAI instructions.
   // If that turns out to be correct, we can probably just remove all the special handling...

   Rotate = No; Flip_X = No; Flip_Y = No; Switches = 0; Inputs = 0xFFFF; vgSetTwinkle(255); useParity=1;

   if (strcasecmp (argv[1], "tailgunner") == 0) {
     Game = GAME_TAILGUNNER;
     Rotate=Yes; Flip_X=Yes; Flip_Y=Yes;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=1;
     JMI=No;
     RomImages[0] = "roms/tailgunner/tailg.t7";
     RomImages[1] = "roms/tailgunner/tailg.p7";
     RomImages[2] = "roms/tailgunner/tailg.u7";
     RomImages[3] = "roms/tailgunner/tailg.r7";
     Switches=0b0000000;
     Inputs=0xBFFF; // 8 way digital joystick mode
     vgSetCineSize(0,0,1024,768); // MinX, MinY, MaxX, MaxY
     //  Monitor=BiLevel
     vgSetTwinkle(7);
   } else if (strcasecmp (argv[1], "ripoff") == 0) {
     Game = GAME_RIPOFF;
     Rotate=Yes; Flip_X=Yes; Flip_Y=Yes;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=1;
     JMI=Yes;
     RomImages[0]="roms/ripoff/ripoff.t7";
     RomImages[1]="roms/ripoff/ripoff.p7";
     RomImages[2]="roms/ripoff/ripoff.u7";
     RomImages[3]="roms/ripoff/ripoff.r7";
     /*
DIP_START(ripoff_dips)
 DIP_SET4(DIPSW1,0xfc,"Units",0x00,"8",0x01,"16",0x02,"4",0x03,"12",3)
 DIP_SET4(DIPSW1,0xf3,"Coins",0x08,"2Coins 1Credit",0x00,"4Coins 3Credits",0x0c,"1Coin 1Credit",0x04,"2Coins 3Credits",2)
 DIP_SET2(DIPSW1,0xef,"Attract Sound",0x10,"Off",0x00,"On",0)
 DIP_SET2(DIPSW1,0xdf,"Scores",0x20,"Individual",0x00,"Combined",0)
 DIP_SET2(DIPSW1,0xbf,"Diagnostic",0x00,"Off",0x40,"On",0)
DIP_END
      */
     Switches=0b1011101;
     //         ^ trying 1 to remove diagnostic frame
     vgSetCineSize(-16,-16,1041,785); // MinX, MinY, MaxX, MaxY
     // Monitor=BiLevel
   } else if (strcasecmp (argv[1], "spacewars") == 0) {
     Game = GAME_SPACEWARS;
     Rotate=Yes; Flip_X=Yes; Flip_Y=Yes;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=0;
     JMI=No;
     RomImages[0]="roms/spacewars/spacewar.u7";
     RomImages[1]="roms/spacewars/spacewar.r7";
     Switches=0b0000011;
     vgSetCineSize(0,0,1024,768); // MinX, MinY, MaxX, MaxY
     // Monitor=BiLevel
     vgSetTwinkle(9);
   } else if (strcasecmp (argv[1], "boxingbugs") == 0) {
     Game = GAME_BOXINGBUGS;
     Rotate=Yes; Flip_X=Yes; Flip_Y=Yes;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=3;
     JMI=Yes;
     // Mouse=BoxingBugs
     // MouseSpeedX=1024	; use negative number to reverse mouse control
     RomImages[0]="roms/boxingbugs/u1a";
     RomImages[1]="roms/boxingbugs/u1b";
     RomImages[2]="roms/boxingbugs/u2a";
     RomImages[3]="roms/boxingbugs/u2b";
     RomImages[4]="roms/boxingbugs/u3a";
     RomImages[5]="roms/boxingbugs/u3b";
     RomImages[6]="roms/boxingbugs/u4a";
     RomImages[7]="roms/boxingbugs/u4b";
     Switches=0b1101100;
     //         ^ that '1' makes it play the game as opposed to a diag mode
     vgSetCineSize(-1,0,1026,784); // MinX, MinY, MaxX, MaxY
     // Monitor=Color
   } else if (strcasecmp (argv[1], "armorattack") == 0) {
     Game = GAME_ARMORATTACK;
     Rotate=Yes; Flip_X=Yes; Flip_Y=Yes;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=2;
     JMI=Yes;
     RomImages[0] = "roms/armorattack/ar414le.t6";
     RomImages[1] = "roms/armorattack/ar414lo.p6";
     RomImages[2] = "roms/armorattack/ar414ue.u6";
     RomImages[3] = "roms/armorattack/ar414uo.r6";
     Switches=0b1000000; // diagnostics, no sound, 1 credit per quarter, 5 jeeps per game ???
     vgSetCineSize(2,2,1023,767); // MinX, MinY, MaxX, MaxY
     // Monitor=BiLevel
   } else if (strcasecmp (argv[1], "starcastle") == 0) {
     Game = GAME_STARCASTLE;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=1;
     JMI=Yes;
     RomImages[0]="roms/starcastle/starcas.t7";
     RomImages[1]="roms/starcastle/starcas.p7";
     RomImages[2]="roms/starcastle/starcas.u7";
     RomImages[3]="roms/starcastle/starcas.r7";
     Switches=0b1000011;
     vgSetCineSize(-9,3,1033,795); // MinX, MinY, MaxX, MaxY
     // Monitor=BiLevel
     vgSetTwinkle(8);
   } else if (strcasecmp (argv[1], "starhawk") == 0) {
     Game = GAME_STARHAWK;
     Rotate=Yes; Flip_X=Yes; Flip_Y=Yes;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=0;
     JMI=Yes;
     RomImages[0]="roms/starhawk/starhawk.u7";
     RomImages[1]="roms/starhawk/starhawk.r7";
     vgSetCineSize(-16,-16,1040,784);
     // Monitor=BiLevel
     vgSetTwinkle(9);
   } else if (strcasecmp (argv[1], "speedfreak") == 0) {
     Game = GAME_SPEEDFREAK;
     Rotate=Yes; Flip_X=Yes; Flip_Y=Yes;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=1;
     JMI=No;
     // Mouse=SpeedFreak
     // MouseSpeedX=1536
     RomImages[0]="roms/speedfreak/speedfrk.t7";
     RomImages[1]="roms/speedfreak/speedfrk.p7";
     RomImages[2]="roms/speedfreak/speedfrk.u7";
     RomImages[3]="roms/speedfreak/speedfrk.r7";
     Inputs=0xFFE0; // zero steering wheel, start in 1st gear
     vgSetCineSize(-9,3,1033,795); // MinX, MinY, MaxX, MaxY
     // Monitor=BiLevel
     vgSetTwinkle(8);
     //Brightness=100,100,100		; RGB percentages for brightest level
     //Contrast=67,67,67		; RGB percentages for dimest level
   } else if (strcasecmp (argv[1], "demon") == 0) {
     Game = GAME_DEMON;
     Rotate=Yes; Flip_X=Yes; Flip_Y=Yes;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=2;
     JMI=Yes;
     RomImages[0]="roms/demon/demon.7t";
     RomImages[1]="roms/demon/demon.7p";
     RomImages[2]="roms/demon/demon.7u";
     RomImages[3]="roms/demon/demon.7r";
     Switches=0b1000000;
     vgSetCineSize(7,7,1021,805); // MinX, MinY, MaxX, MaxY
     // Monitor=BiLevel
     vgSetTwinkle(8);
   } else if (strcasecmp (argv[1], "solarquest") == 0) {
     Game = GAME_SOLARQUEST;
     //if (settings->orientation == 0) {
       Rotate=Yes; Flip_Y=Yes; // hardware correction in X
       //}
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=2;
     JMI=Yes;
     RomImages[0]="roms/solarquest/solar.t7";
     RomImages[1]="roms/solarquest/solar.p7";
     RomImages[2]="roms/solarquest/solar.u7";
     RomImages[3]="roms/solarquest/solar.r7";
     //Switches=0b1111000;
     Switches=0;
     vgSetCineSize(0,0,1020,768); // MinX, MinY, MaxX, MaxY
     // Monitor=64Level
   } else if (strcasecmp (argv[1], "cosmicchasm") == 0) {
     Game = GAME_COSMICCHASM;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=3; // this may be a 64k rom
     JMI=Yes;
     // Mouse=BoxingBugs
     // MouseSpeedX=1024	; use negative number to reverse mouse control
     RomImages[0]="roms/cosmicchasm/chasm.u4";
     RomImages[1]="roms/cosmicchasm/chasm.u12";
     RomImages[2]="roms/cosmicchasm/chasm.u8";
     RomImages[3]="roms/cosmicchasm/chasm.u16";
     RomImages[4]="roms/cosmicchasm/chasm.u3";
     RomImages[5]="roms/cosmicchasm/chasm.u11";
     RomImages[6]="roms/cosmicchasm/chasm.u7";
     RomImages[7]="roms/cosmicchasm/chasm.u15";
     Switches=0b0101100;
     vgSetCineSize(-1,0,1026,784); // MinX, MinY, MaxX, MaxY
     // Monitor=Color
   } else if (strcasecmp (argv[1], "waroftheworlds") == 0) {
     Game = GAME_WAROFTHEWORLDS;
     Rotate=Yes; Flip_X=Yes; Flip_Y=Yes;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=2;
     JMI=Yes;
     RomImages[0]="roms/waroftheworlds/wotw.t7";
     RomImages[1]="roms/waroftheworlds/wotw.p7";
     RomImages[2]="roms/waroftheworlds/wotw.u7";
     RomImages[3]="roms/waroftheworlds/wotw.r7";
     //ORIG: Switches=0b0100000; // free play
     Switches=0b01000000; // debugging diag mode startup
     vgSetCineSize(-27,0,1060,768); // MinX, MinY, MaxX, MaxY
     // Monitor=Color
   } else if (strcasecmp (argv[1], "warrior") == 0) {
     Game = GAME_WARRIOR;
     Rotate=Yes; Flip_X=Yes; Flip_Y=Yes;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=1;
     JMI=Yes;
     RomImages[0]="roms/warrior/warrior.t7";
     RomImages[1]="roms/warrior/warrior.p7";
     RomImages[2]="roms/warrior/warrior.u7";
     RomImages[3]="roms/warrior/warrior.r7";
     //Switches=0b1111111;
     Switches=0b00000000;
     vgSetCineSize(0,0,1024,768); // MinX, MinY, MaxX, MaxY
     // Monitor=BiLevel
   } else if (strcasecmp (argv[1], "barrier") == 0) {
     Game = GAME_BARRIER;
     //Rotate=Yes;
     //if (settings->orientation == 0) {
       Flip_X=Yes; Flip_Y=Yes;
       //}
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=0;
     JMI=Yes;
     RomImages[0]="roms/barrier/barrier.t7";
     RomImages[1]="roms/barrier/barrier.p7";
     Switches=0b0000001;
     vgSetCineSize(67,74,981,693); // MinX, MinY, MaxX, MaxY
     // Monitor=BiLevel
   } else if (strcasecmp (argv[1], "sundance") == 0) {
     Game = GAME_SUNDANCE;
     //Rotate=Yes;
     //if (settings->orientation != 0) {
       Flip_X=Yes; Flip_Y=Yes;
       //}
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=1;
     JMI=Yes;
     RomImages[0]="roms/sundance/sundance.t7";
     RomImages[1]="roms/sundance/sundance.p7";
     RomImages[2]="roms/sundance/sundance.u7";
     RomImages[3]="roms/sundance/sundance.r7";
     Switches=0b0000101;
     vgSetCineSize(0,0,1024,768); // MinX, MinY, MaxX, MaxY
     // Monitor=16Level
     vgSetTwinkle(0);
   } else if (strcasecmp (argv[1], "qb3") == 0) {
     Game = GAME_QB3;
     // 0=4k, 1=8k, 2=16k, 3=32k
     MemSize=4; // only game with 4 x 8k roms...
     JMI=No;
     RomImages[0]="roms/qb3/qb3_le_t7.bin";
     RomImages[1]="roms/qb3/qb3_lo_p7.bin";
     RomImages[2]="roms/qb3/qb3_ue_u7.bin";
     RomImages[3]="roms/qb3/qb3_uo_r7.bin";
     // Inputs unknown for now # <input players="1" buttons="4" coins="1">
     Switches=0b1001001000010; // free play on.  Is this extended switch plausible???
     vgSetCineSize(-9,3,1033,795); // MinX, MinY, MaxX, MaxY
     // Monitor=BiLevel ; probably not.  should research this game.
     vgSetTwinkle(8);
     //Brightness=100,100,100		; RGB percentages for brightest level
     //Contrast=67,67,67		; RGB percentages for dimmest level
   } else {
      fprintf (stderr, "The supported games are:\n");
      fprintf (stderr, "  armorattack  boxingbugs   demon  ripoff      spacewars   starcastle  sundance    waroftheworlds\n");
      fprintf (stderr, "  barrier      cosmicchasm  qb3    solarquest  speedfreak  starhawk    tailgunner  warrior\n");
      fprintf (stderr, "\n");
      exit (1);
   }

   cineSetGame (argv[1], Game);

   if (readRoms (RomImages, MemSize) != 0) exit (1);

   cineSetJMI (JMI);		 // set JMI option
   cineSetMSize (MemSize);	 // set memory size
   cineSetRate (38);             // Ini files support "Standard" vs "Full" but not used.
   cineSetSw (Switches, Inputs); // set switches & inputs
   // standalone mode already has rotation applied...
   //if (settings->orientation == 0) {
     // By now screen is correctly rotated for horizontal mode
     // If menu orientation is 'rotated' (assume 90 anti-clockwise as per standard stand)
     // then unrotate back to vertical mode.
     Rotate = !Rotate;
     Flip_X = !Flip_X;
     Flip_Y = !Flip_Y;
   //}
   vgSetRotateFlip (Rotate, Flip_X, Flip_Y);
   vgInit ();

   // #################################################################################
   // ##                                                                             ##
   // ##                               EMULATE!                                      ##
   // ##                                                                             ##
   // #################################################################################
   cineInit(); 	 // setup segment pointer, reset breakpoints
   cineReset();	 // called after all initialisations complete

   startFrame(); // needs an initial call to get start values of buttons, cursor etc
                 // - it is *start* Frame after all, not *end* Frame...

   for (;;) {
     cineExec (); // run an unspecified number of CCPU instructions
     if (bNewFrame) { // If a complete frame's worth ..?
       if (useParity) {
         parity ^= 1;
         if (parity) startFrame(); // this will call game-specific I/O handler
       } else {
         startFrame();
       }
     }
   }   
   exit(0); // probably never reached.
   return 0; // definitely never reached.
}
