/*
    Use this to generate a 'default' config under Raspbian, since there is no
   menu system to create a top-level config file.

   We draw a rectangle between (-18000,-24000) and (18000,24000) in default vertical
   mode, and ask the user to adjust the display positioning so that the outline of
   that rectangle is right on the edge of the display, partially visible on all 4 sides.
   These are coordinates as passed to v_directDraw32 etc, ie not scaled by the window
   handling.

   All games will then assume that these are actual device coordinates and will position
   the virtual window centrally within that rectangle.  Any positioning of the game's
   contents within the virtual window should be tweaked within the game, NOT by using
   the vectrex positioning menu on a per-game basis.  This way the config tweaking only
   needs to be done once, not once per game or worse twice per game, with once for
   vertical mode and again for horizontal mode.
 */

#include <stdio.h>
#include <pitrex/pitrexio-gpio.h>
#include <vectrex/vectrexInterface.h>

#ifndef TRUE
#define TRUE (0==0)
#define FALSE (!TRUE)
#endif

void startFrame()
{
  v_WaitRecal();
  //v_doSound();
  v_setBrightness(64);        /* set intensity of vector beam... */
  v_readButtons();
  v_readJoystick1Analog();
  //v_playAllSFX();
}

// Windows handling done by local code.  Windows.c is the master copy. Eventually this
// will migrate to the vectrex library.

static int64_t ScaleXMul=1LL, ScaleXDiv=1LL, ScaleXOffsetPre=0LL, ScaleXOffsetPost=0LL,
               ScaleYMul=1LL, ScaleYDiv=1LL, ScaleYOffsetPre=0LL, ScaleYOffsetPost=0LL;

int tx(int x) { // convert x from window to viewport
  return (int)(((((int64_t)x)+ScaleXOffsetPre)*ScaleXMul)/ScaleXDiv + ScaleXOffsetPost);
}

int ty(int y) { // and y
  return (int)(((((int64_t)y)+ScaleYOffsetPre)*ScaleYMul)/ScaleYDiv + ScaleYOffsetPost);
}

void window(int xl, int yb, int xr, int yt) {
  // Does not yet support flip/rotate
  // We will use normalised viewport coordinates of x: -18000 : 18000 and y: -24000 : 24000
  int64_t width, height;
  int xc, yc;
  int oxl = xl, oyb = yb, oxr = xr, oyt = yt;
  width = (int64_t)xr-(int64_t)xl;
  height = (int64_t)yt-(int64_t)yb;
  if (width*4 > height*3) {
    // window is wider than aspect ratio, so we will have black bars at the top and bottom
    height = (width * 4) / 3;
    yc = (yb+yt)/2;
    yb = yc - height/2;
    yt = yc + height/2;
  } else if (width*4 < height*3) {
    // window is taller than aspect ratio, so we will have black bars at the sides
    width = (height*3) / 4;
    xc = (xl+xr)/2;
    xl = xc - width/2;
    xr = xc + width/2;
  }
  ScaleXMul = 36000LL; ScaleXDiv = width; ScaleXOffsetPre = -width/2LL; ScaleXOffsetPost = 0LL; ScaleXOffsetPost = (tx(xr) - tx(oxr)) / 2LL;
  ScaleYMul = 48000LL; ScaleYDiv = height; ScaleYOffsetPre = -height/2LL; ScaleYOffsetPost = 0LL; ScaleYOffsetPost = (ty(yt) - ty(oyt)) / 2LL;
  // setCustomClipping(TRUE, tx(oxl), ty(oyb), tx(oxr), ty(oyt));
  // transform world (window) coordinates to viewport (normalised device coordinates) before
  // clipping.  That way clipping code does not need to know about world coordinates.
  // not implemented as I don't have the updated library yet
  // Note that we don't always want to clip, because we might want to move some of the
  // elements of a game's display, for instance if it was a horizontal mode game and
  // we're converting it to vertical mode, then we might want to place something such
  // as the score or the copyright message outside of the original window rectangle.
}

void line(int xl, int yb, int xr, int yt, int col) {
  //fprintf(stdout, "line(%d,%d, %d,%d, %d);n", tx(xl),ty(yb), tx(xr),ty(yt), col);
  v_directDraw32(tx(xl),ty(yb), tx(xr),ty(yt), col);
}

int main(int argc, char **argv) {
  int last_setting = -1;
  vectrexinit(1);
  v_init();
  usePipeline = 1;
  v_setRefresh(60);
  for (;;) {
    if (settings->orientation != last_setting) {
      if ((settings->orientation&1) == 0) {
        window(0,0, 360,480); // same aspect ratio 3:4
      } else {
        window(0,0, 360,480);
      }
      last_setting = settings->orientation;
    }
    startFrame();
    line(  0,   0,   0, 480, 96);
    line(  0, 480, 360, 480, 96);
    line(360, 480, 360,   0, 96);
    line(360,   0,   0,   0, 96);
    line(  80, 140,  80,  340, 64);
    line(  80, 340, 280,  340, 64);
    line( 280, 340, 280,  140, 64);
    line( 280, 140,  80,  140, 64);
    v_printString(0,0, (settings->orientation == 0 ? "VERTICAL" :
		       (settings->orientation == 1 ? "HORIZONTAL" :
		       (settings->orientation == 2 ? "INVERTED VERTICAL" :
		                                     "INVERTED HORIZONTAL"))), 7, 64);
  }
  return 0;
}
